#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

def calc_image_cell_size(vis_xds, min_dish_diameter):
    """
    Calculates an appropriate number of pixels and and cell size for imaging a measurement set.
    It uses the perfectly-illuminated circular aperture approximation to determine the field of view
    and 7 pixel per beam for the cell size.

    Parameters
    ----------
    vis_xds : xarray.core.dataset.Dataset
        input Visibility Dataset
    min_dish_diameter : float
        smallest antenna diameter
    Returns
    -------
    imsize : list of ints
        number of pixels for each spatial dimension. 
    cell : list of ints
        Cell size is arcseconds.
    """
    import xarray
    import numpy as np
    import dask.array  as da
    rad_to_arc = (3600 * 180) / np.pi  # Radians to arcseconds
    c = 299792458

    f_min = da.nanmin(vis_xds.chan)
    f_max = da.nanmax(vis_xds.chan)
    # D_min = np.nanmin(global_xds.ANT_DISH_DIAMETER)
    D_min = min_dish_diameter

    # Calculate cell size using 7 pixels per beam
    cell = rad_to_arc * np.array(
        [c / (da.nanmax(vis_xds.UVW[:, :, 0].data) * f_max), c / (da.nanmax(vis_xds.UVW[:, :, 1].data) * f_max)]) / 7

    # If cell sizes are within 20% of each other use the smaller cell size for both.
    if (cell[0] / cell[1] < 1.2) and (cell[1] / cell[0] < 1.2):
        cell[:] = np.min(cell)

    # Calculate imsize using the perfectly-illuminated circular aperture approximation
    FWHM_max = np.array((rad_to_arc * (1.02 * c / (D_min * f_min))))
    imsize = FWHM_max / cell

    # Find an image size that is (2^n)*10 when muliplied with the gridding padding and n is an integer
    padding = 1.2

    if imsize[0] < 1:
        imsize[0] = 1

    if imsize[1] < 1:
        imsize[1] = 1

    n_power = np.ceil(np.log2(imsize / 10))
    imsize = np.ceil(((2 ** n_power) * 10) / padding)

    return cell, imsize



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


def standard_grid(grid_data, uvw, weight, flag_row, flag, freq_chan, chan_map, n_chan, pol_map, cgk_1D, grid_parms):
    """
    Grids visibilities.

    Parameters
    ----------
    grid : complex array (n_chan, n_pol, n_u, n_v)
    sum_weight : float array (n_chan, n_pol)
    uvw  : float array (n_time, n_baseline, 3)
    freq_chan : float array (n_chan)
    chan_map : int array (n_chan)
    pol_map : int array (n_pol)
    weight : float array (n_time, n_baseline, n_vis_chan)
    flag_row : boolean array (n_time, n_baseline)
    flag : boolean array (n_time, n_baseline, n_chan, n_pol)
    cgk_1D : float array (oversampling*(support//2 + 1))
    grid_parms : dictionary ('n_imag_chan','n_imag_pol','n_uv','delta_lm','oversampling','support')
    Returns
    -------
    grid : complex array (n_imag_chan,n_imag_pol,n_u,n_v)
    sum_weight : float array (n_imag_chan,n_imag_pol)
    """
    import numpy as np
    import dask.array as da
    from dask.diagnostics import ProgressBar

    n_imag_chan = grid_parms['n_imag_chan']
    n_imag_pol = grid_parms['n_imag_pol']
    n_uv = grid_parms['n_uv']
    oversampling = grid_parms['oversampling']
    support = grid_parms['support']
    sum_weight = np.ones((n_imag_chan, n_imag_pol), dtype=np.double)
    c = 299792458.0

    # everything needs to be ordered by the same dimensions, too confusing otherwise
    # (time x baseline x channel x pol x support^2 x uv)
    # use singleton dimensions as placeholders (np.newaxis / None) so size doesn't change
    # this allows easy broadcasting for matrix math later on
    chunks = (200,100,1,1,50,-1)
    mask = da.from_array(flag_row[..., None,None,None,None] | flag[...,None,None], chunks=chunks)
    weighted_data = da.from_array(grid_data[...,None,None] * weight[...,None,:,None,None], chunks=chunks)
    support_idxs = (np.mgrid[0:support, 0:support] - support // 2).reshape(2,-1).T[None,None,None,None,...]
    
    uv_scale = da.from_array((freq_chan[:,None,None,None] * (grid_parms['delta_lm']*n_uv/c))[None,None,...], chunks=chunks)
    uv = da.from_array(-uvw[...,None,None,None,:2], chunks=chunks) * uv_scale
    uv_pos = uv + (n_uv // 2)
    
    uv_center_indx = (uv_pos + 0.5).astype(int)
    uv_indx = (support_idxs + uv_center_indx)
    uv_center_offset_indx = da.floor((uv_center_indx - uv_pos) * oversampling + 0.5).astype(int)
    uv_offset_indx = da.fabs(oversampling * support_idxs + uv_center_offset_indx).astype(int)
    
    # TESTING
    # weighted_data_c = weighted_data.blocks[0,0,100,0,0,0].compute()
    # uv_offset_indx_c = uv_offset_indx.blocks[0,0,100,0,0,0].compute()
    # uv_indx_c = uv_indx.blocks[0,0,100,0,0,0].compute()
    # mask_c = mask.blocks[0,0,100,0,0,0].compute()
    
    # map the gridding across chunks
    # noinspection PyTypeChecker
    def set_grid(weighted_data_c, uv_offset_indx_c, uv_indx_c, mask_c, support, n_uv):
        mask_c = np.where(mask_c.repeat(support, 4).ravel() == False)  # turn in to a filter
        conv = cgk_1D[uv_offset_indx_c.clip(0,len(cgk_1D)-1)].prod(axis=5, keepdims=True)
        gvals_1d = (conv * weighted_data_c).ravel()[mask_c]
        gcoords_1d = np.ravel_multi_index(uv_indx_c.reshape(-1, 2).T, n_uv, mode='clip')[mask_c]
        #norm = np.sum(conv)
        
        grid_chunk1d = np.bincount(gcoords_1d, np.real(gvals_1d), minlength=np.prod(n_uv))
        grid_chunk1d = grid_chunk1d + 1j * np.bincount(gcoords_1d, np.imag(gvals_1d), minlength=np.prod(n_uv))
        grid_chunk = grid_chunk1d.reshape(tuple(n_uv) + tuple([1 for _ in range(6)]))
        return grid_chunk
    
    
    da_grid = da.map_blocks(set_grid, weighted_data, uv_offset_indx, uv_indx, mask, support, n_uv,
                            dtype=weighted_data.dtype, chunks=tuple(n_uv)+weighted_data.chunksize)
    da_grid = da_grid.sum(axis=[2,3,6,7], keepdims=False)
    
    with ProgressBar():
        grid = da_grid.compute()
    
    grid = grid.sum(axis=3).sum(axis=2)
    
    #sum_weight[a_chan, a_pol] = sum_weight[a_chan, a_pol] + weight[i_time, i_baseline, i_pol] * norm
    #
    #
    #grid = grid / sum_weight[:, :, None, None]
    return grid, sum_weight



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

def dirty_image(vis_dataset, grid_parms):
    """
    Grids visibilities from Visibility Dataset and returns dirty Image Dataset.
    If to_disk is set to true the data is saved to disk.
    Parameters
    ----------
    vis_xds : xarray.core.dataset.Dataset
        input Visibility Dataset
    grid_parms : dictionary
          keys ('chan_mode','imsize','cell','oversampling','support','to_disk','outfile')
    Returns
    -------
    dirty_image_xds : xarray.core.dataset.Dataset

    """

    import numpy as np
    from numba import jit
    import time
    import math
    import dask.array.fft as dafft
    import xarray as xr
    import dask.array as da
    import matplotlib.pylab as plt
    from .grid import _graph_grid
    import dask.array.fft as dafft
    import dask
    import copy, os
    from numcodecs import Blosc
    from itertools import cycle

    # Parameter adjustments
    grid_parms_dirty_image = copy.deepcopy(grid_parms)
    padding = 1.2  # Padding factor
    dtr = np.pi / (3600 * 180)
    grid_parms_dirty_image['imsize'] = (padding * np.array(grid_parms_dirty_image['imsize'])).astype(int)  # Add padding
    grid_parms_dirty_image['cell'] = np.array(grid_parms_dirty_image['cell']) * dtr
    grid_parms_dirty_image['cell'][0] = -grid_parms_dirty_image['cell'][0]
    grid_parms_dirty_image['to_disk'] = False

    assert grid_parms_dirty_image['chan_mode'] == 'continuum' or grid_parms_dirty_image[
        'chan_mode'] == 'cube', 'The chan_mode parameter in grid_parms can only be \'continuum\' or \'cube\'.'
    
    grids_and_sum_weights, correcting_cgk_image = _graph_grid(vis_dataset, grid_parms_dirty_image)
    

    # uncorrected_dirty_image = dafft.fftshift(dafft.ifft2(dafft.ifftshift(grids_and_sum_weights[0], axes=(2,3)), axes=(2,3)), axes=(2,3))
    # uncorrected_dirty_image = uncorrected_dirty_image.real * (grid_parms['imsize'][0] * grid_parms['imsize'][1])
    uncorrected_dirty_image = dafft.fftshift(
        dafft.ifft2(dafft.ifftshift(grids_and_sum_weights[0], axes=(0, 1)), axes=(0, 1)), axes=(0, 1))
    uncorrected_dirty_image = uncorrected_dirty_image.real * (
            grid_parms_dirty_image['imsize'][0] * grid_parms_dirty_image['imsize'][1])

    def correct_image(uncorrected_dirty_image, sum_weights, correcting_cgk):
        sum_weights[sum_weights == 0] = 1
        # corrected_image = (uncorrected_dirty_image/sum_weights[:,:,None,None])/correcting_cgk[None,None,:,:]
        corrected_image = (uncorrected_dirty_image / sum_weights[None, None, :, :]) / correcting_cgk[:, :, None, None]
        return corrected_image

    corrected_dirty_image = da.map_blocks(correct_image, uncorrected_dirty_image, grids_and_sum_weights[1],
                                          correcting_cgk_image)  # ? has to be .data to paralize correctly
   
    #Remove Padding
    print(grid_parms)
    n_xy = (np.array(grid_parms['imsize'])).astype(np.int)
    n_xy_padded = (padding*n_xy).astype(int)
    start_xy = (n_xy_padded // 2 - n_xy // 2)
    end_xy = start_xy + n_xy

    if grid_parms_dirty_image['chan_mode'] == 'continuum':
        freq_coords = [da.mean(vis_dataset.coords['chan'].values)]
        imag_chan_chunk_size = 1
    elif grid_parms_dirty_image['chan_mode'] == 'cube':
        freq_coords = vis_dataset.coords['chan'].values
        imag_chan_chunk_size = vis_dataset.DATA.chunks[2][0]

    chunks = vis_dataset.DATA.chunks
    n_imag_pol = chunks[3][0]
    dirty_image_dict = {}
    coords = {'d0': np.arange(grid_parms['imsize'][0]), 'd1': np.arange(grid_parms['imsize'][1]),
              'chan': freq_coords, 'pol': np.arange(n_imag_pol)}
    dirty_image_dict['CORRECTING_CGK'] = xr.DataArray(da.array(correcting_cgk_image[start_xy[0]:end_xy[0], start_xy[1]:end_xy[1]]), dims=['d0', 'd1'])
    # dirty_image_dict['VIS_GRID'] = xr.DataArray(grids_and_sum_weights[0], dims=['chan','pol','u', 'v'])
    #dirty_image_dict['VIS_GRID'] = xr.DataArray(grids_and_sum_weights[0], dims=['d0', 'd1', 'chan', 'pol'])
    dirty_image_dict['SUM_WEIGHT'] = xr.DataArray(grids_and_sum_weights[1], dims=['pol','chan'])
    dirty_image_dict['DIRTY_IMAGE'] = xr.DataArray(corrected_dirty_image[start_xy[0]:end_xy[0], start_xy[1]:end_xy[1],:,:], dims=['d0', 'd1', 'pol', 'chan'])
    dirty_image_xds = xr.Dataset(dirty_image_dict, coords=coords)
    
    if grid_parms['to_disk'] == True:
        outfile = grid_parms['outfile']
        tmp = os.system("rm -fr " + outfile)
        tmp = os.system("mkdir " + outfile)

        compressor = Blosc(cname='zstd', clevel=2, shuffle=0)
        encoding = dict(zip(list(dirty_image_xds.data_vars), cycle([{'compressor': compressor}])))
        start = time.time()
        xr.Dataset.to_zarr(dirty_image_xds, store=outfile, mode='w', encoding=encoding)
        dirty_image_time = time.time() - start
        print('Dirty Image time ', time.time() - start)

        dirty_image_xds = xr.open_zarr(outfile)
        dirty_image_xds.attrs['dirty_image_time'] = dirty_image_time
        return dirty_image_xds
    else:
        return dirty_image_xds



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from numba import jit
import numpy as np
import math

def grid(vis_dataset, grid_parms):
    """
    Grids visibilities from Visibility Dataset.
    If to_disk is set to true the data is saved to disk.
    Parameters
    ----------
    vis_xds : xarray.core.dataset.Dataset
        input Visibility Dataset
    grid_parms : dictionary
          keys ('chan_mode','imsize','cell','oversampling','support','to_disk','outfile')
    Returns
    -------
    grid_xds : xarray.core.dataset.Dataset

    """
    from numcodecs import Blosc
    import os
    from itertools import cycle
    import dask.array as da
    import xarray as xr
    import dask
    import time
    import copy

    # Parameter adjustments
    grid_parms_copy = copy.deepcopy(grid_parms)
    padding = 1.2  # Padding factor
    dtr = np.pi / (3600 * 180)
    grid_parms_copy['imsize'] = (padding * np.array(grid_parms_copy['imsize'])).astype(int)  # Add padding
    grid_parms_copy['cell'] = np.array(grid_parms_copy['cell']) * dtr
    grid_parms_copy['cell'][0] = -grid_parms_copy['cell'][0]

    assert grid_parms_copy['chan_mode'] == 'continuum' or grid_parms_copy[
        'chan_mode'] == 'cube', 'The chan_mode parameter in grid_parms can only be \'continuum\' or \'cube\'.'

    if grid_parms_copy['chan_mode'] == 'continuum':
        freq_coords = [da.mean(vis_dataset.coords['chan'].values)]
        imag_chan_chunk_size = 1
    elif grid_parms_copy['chan_mode'] == 'cube':
        freq_coords = vis_dataset.coords['chan'].values
        imag_chan_chunk_size = vis_dataset.DATA.chunks[2][0]

    # Create dask graph of gridding
    grids_and_sum_weights, correcting_cgk_image = _graph_grid(vis_dataset, grid_parms_copy)

    # Create delayed xarray dataset
    chunks = vis_dataset.DATA.chunks
    n_imag_pol = chunks[3][0]
    grid_dict = {}
    coords = {'chan': freq_coords, 'pol': np.arange(n_imag_pol), 'u': np.arange(grid_parms_copy['imsize'][0]),
              'v': np.arange(grid_parms_copy['imsize'][1])}
    grid_dict['CORRECTING_CGK'] = xr.DataArray(da.array(correcting_cgk_image), dims=['u', 'v'])
    # grid_dict['VIS_GRID'] = xr.DataArray(grids_and_sum_weights[0], dims=['chan','pol','u', 'v'])
    grid_dict['VIS_GRID'] = xr.DataArray(grids_and_sum_weights[0], dims=['u', 'v', 'pol', 'chan'])
    grid_dict['SUM_WEIGHT'] = xr.DataArray(grids_and_sum_weights[1], dims=['pol','chan'])
    grid_xds = xr.Dataset(grid_dict, coords=coords)

    if grid_parms['to_disk'] == True:
        outfile = grid_parms['outfile']
        tmp = os.system("rm -fr " + outfile)
        tmp = os.system("mkdir " + outfile)

        compressor = Blosc(cname='zstd', clevel=2, shuffle=0)
        encoding = dict(zip(list(grid_xds.data_vars), cycle([{'compressor': compressor}])))
        start = time.time()
        xr.Dataset.to_zarr(grid_xds, store=outfile, mode='w', encoding=encoding)
        grid_time = time.time() - start
        print('Grid time ', time.time() - start)

        grid_xds = xr.open_zarr(outfile)
        grid_xds.attrs['grid_time'] = grid_time
        return grid_xds

    else:
        return grid_xds


def _graph_grid(vis_dataset, grid_parms):
    import dask
    import dask.array as da
    import xarray as xr
    import time
    import itertools
    from cngi.gridding import gridding_convolutional_kernels as gck
    

    ##############################
    # Chunking on Polarization not supported add assert statement
    # Smeered poinst outside grid are ignored
    # Remove pol chunking option (pols are inhereantly connected). Should always be in same block
    # Convert weight to weight spectrum and keep (weight name)
    ##############################

    # Creating gridding kernel
    # The support in CASA is defined as the half support, which is 3
    cgk, correcting_cgk_image = gck.create_prolate_spheroidal_kernel(grid_parms['oversampling'], grid_parms['support'],
                                                                     grid_parms['imsize'])
    cgk_1D = gck.create_prolate_spheroidal_kernel_1D(grid_parms['oversampling'], grid_parms['support'])

    # Getting data for gridding
    chan_chunk_size = vis_dataset.DATA.chunks[2][0]

    freq_chan = da.from_array(vis_dataset.coords['chan'].values, chunks=(chan_chunk_size))

    n_chunks_in_each_dim = vis_dataset.DATA.data.numblocks
    chunk_indx = []

    iter_chunks_indx = itertools.product(np.arange(n_chunks_in_each_dim[0]), np.arange(n_chunks_in_each_dim[1]),
                                         np.arange(n_chunks_in_each_dim[2]), np.arange(n_chunks_in_each_dim[3]))

    assert grid_parms['chan_mode'] == 'continuum' or grid_parms[
        'chan_mode'] == 'cube', 'The chan_mode parameter in grid_parms can only be \'continuum\' or \'cube\'.'

    if grid_parms['chan_mode'] == 'continuum':
        n_chan_chunks_img = 1
    elif grid_parms['chan_mode'] == 'cube':
        n_chan_chunks_img = n_chunks_in_each_dim[2]

    list_of_sub_grids_and_sum_weights = [[] for i in range(n_chan_chunks_img)]
    n_delayed = np.prod(n_chunks_in_each_dim)
    chunk_sizes = vis_dataset.DATA.chunks

 
    list_of_grids = [[] for i in range(n_chan_chunks_img)]
    list_of_sum_weights = [[] for i in range(n_chan_chunks_img)]
    # Build graph
    for c_time, c_baseline, c_chan, c_pol in iter_chunks_indx:
        sub_grid_and_sum_weights = dask.delayed(_standard_grid_dask)(
            vis_dataset.DATA.data.partitions[c_time, c_baseline, c_chan, c_pol],
            vis_dataset.UVW.data.partitions[c_time, c_baseline, 0],
            vis_dataset.WEIGHT.data.partitions[c_time, c_baseline, c_pol],
            vis_dataset.FLAG_ROW.data.partitions[c_time, c_baseline],
            vis_dataset.FLAG.data.partitions[c_time, c_baseline, c_chan, c_pol],
            freq_chan.partitions[c_chan],
            dask.delayed(cgk_1D), dask.delayed(grid_parms))
        if grid_parms['chan_mode'] == 'continuum':
           list_of_grids[0].append(da.from_delayed(sub_grid_and_sum_weights[0], (
           1, chunk_sizes[3][c_pol], grid_parms['imsize'][0], grid_parms['imsize'][1]),
                                                       dtype=np.complex128))
           list_of_sum_weights[0].append(da.from_delayed(sub_grid_and_sum_weights[1],
           (1, chunk_sizes[3][c_pol]),
           dtype=np.float64))
        elif grid_parms['chan_mode'] == 'cube':
            list_of_grids[c_chan].append(da.from_delayed(sub_grid_and_sum_weights[0], (
            chunk_sizes[2][c_chan], chunk_sizes[3][c_pol], grid_parms['imsize'][0], grid_parms['imsize'][1]),
                                                        dtype=np.complex128))
            list_of_sum_weights[c_chan].append(da.from_delayed(sub_grid_and_sum_weights[1],
            (chunk_sizes[2][c_chan], chunk_sizes[3][c_pol]),
            dtype=np.float64))

    # Sum grids
    for c_chan in range(n_chan_chunks_img):
        list_of_grids[c_chan] = tree_sum_list(list_of_grids[c_chan])
        list_of_sum_weights[c_chan] = tree_sum_list(list_of_sum_weights[c_chan])

    # Concatenate Cube
    if grid_parms['chan_mode'] == 'cube':
        list_of_grids_and_sum_weights = [da.concatenate(list_of_grids,axis=1)[0],da.concatenate(list_of_sum_weights,axis=1)[0]]
    else:
        list_of_grids_and_sum_weights = [list_of_grids[0][0],list_of_sum_weights[0][0]]
    
    #print(list_of_grids_and_sum_weights)
    # Put axes in image orientation. How much does this add to compute?
    #list_of_grids_and_sum_weights[0] = da.moveaxis(list_of_grids_and_sum_weights[0], [-2, -1],
    #                                                            [0, 1])
    list_of_grids_and_sum_weights[0] = da.moveaxis(list_of_grids_and_sum_weights[0], [0, 1],
                                                                [-1, -2])
    
    list_of_grids_and_sum_weights[1] = da.moveaxis(list_of_grids_and_sum_weights[1],[0, 1], [-1, -2])
    
    return list_of_grids_and_sum_weights, correcting_cgk_image
    

def tree_sum_list(list_to_sum):
    import dask.array as da
    while len(list_to_sum) > 1:
        new_list_to_sum = []
        for i in range(0, len(list_to_sum), 2):
            if i < len(list_to_sum) - 1:
                lazy = da.add(list_to_sum[i],list_to_sum[i+1])
            else:
                lazy = list_to_sum[i]
            new_list_to_sum.append(lazy)
        list_to_sum = new_list_to_sum
    return list_to_sum

def _standard_grid_dask(vis_data, uvw, weight, flag_row, flag, freq_chan, cgk_1D, grid_parms):
    """
      Testing Function
      Wrapper function that is used when using dask distributed parallelism (blockwise function). 
      
      Parameters
      ----------
      grid : complex array
          (n_chan, n_pol, n_u, n_v)
      sum_weight : float array
          (n_chan, n_pol) 
      uvw  : float array
          (n_time, n_baseline, 3)
      freq_chan : float array
          (n_chan)
      weight : float array
          (n_time, n_baseline, n_vis_chan)
      flag_row : boolean array
          (n_time, n_baseline)
      flag : boolean array
          (n_time, n_baseline, n_chan, n_pol)
      cgk_1D : float array
          (oversampling*(support//2 + 1))
      grid_parms : dictionary
          keys ('imsize','cell','oversampling','support')

      Returns
      -------
      grid : complex array
          (1,n_imag_chan,n_imag_pol,n_u,n_v)
      """

    n_chan = vis_data.shape[2]
    if grid_parms['chan_mode'] == 'cube':
        n_imag_chan = n_chan
        chan_map = (np.arange(0, n_chan)).astype(np.int)
    else:  # continuum
        n_imag_chan = 1  # Making only one continuum image.
        chan_map = (np.zeros(n_chan)).astype(np.int)

    n_imag_pol = vis_data.shape[3]
    pol_map = (np.arange(0, n_imag_pol)).astype(np.int)

    n_uv = grid_parms['imsize']
    delta_lm = grid_parms['cell']
    oversampling = grid_parms['oversampling']
    support = grid_parms['support']

    grid = np.zeros((n_imag_chan, n_imag_pol, n_uv[0], n_uv[1]), dtype=np.complex128)
    sum_weight = np.zeros((n_imag_chan, n_imag_pol), dtype=np.double)

    _standard_grid_jit(grid, sum_weight, vis_data, uvw, freq_chan, chan_map, pol_map, weight, flag_row, flag, cgk_1D,
                       n_uv, delta_lm, support, oversampling)

    return grid, sum_weight


@jit(nopython=True, cache=True, nogil=True)
def _standard_grid_jit(grid, sum_weight, vis_data, uvw, freq_chan, chan_map, pol_map, weight, flag_row, flag, cgk_1D,
                       n_uv, delta_lm, support, oversampling):
    """
      Parameters
      ----------
      grid : complex array 
          (n_chan, n_pol, n_u, n_v)
      sum_weight : float array 
          (n_chan, n_pol) 
      vis_data : complex array 
          (n_time, n_baseline, n_vis_chan, n_pol)
      uvw  : float array 
          (n_time, n_baseline, 3)
      freq_chan : float array 
          (n_chan)
      chan_map : int array 
          (n_chan)
      pol_map : int array 
          (n_pol)
      weight : float array 
          (n_time, n_baseline, n_vis_chan)
      flag_row : boolean array 
          (n_time, n_baseline)
      flag : boolean array 
          (n_time, n_baseline, n_chan, n_pol)
      cgk_1D : float array 
          (oversampling*(support//2 + 1))
      grid_parms : dictionary 
          keys ('n_imag_chan','n_imag_pol','n_uv','delta_lm','oversampling','support')

      Returns
      -------
      """

    c = 299792458.0
    uv_scale = np.zeros((2, len(freq_chan)), dtype=np.double)
    uv_scale[0, :] = (freq_chan * delta_lm[0] * n_uv[0]) / c
    uv_scale[1, :] = (freq_chan * delta_lm[1] * n_uv[1]) / c

    oversampling_center = int(oversampling // 2)
    support_center = int(support // 2)
    uv_center = n_uv // 2

    n_time = uvw.shape[0]
    n_baseline = uvw.shape[1]
    n_chan = len(chan_map)
    n_pol = len(pol_map)
    n_imag_chan = chan_map.shape[0]

    n_u = n_uv[0]
    n_v = n_uv[1]

    for i_time in range(n_time):
        for i_baseline in range(n_baseline):
            if flag_row[i_time, i_baseline] == 0:
                for i_chan in range(n_chan):
                    a_chan = chan_map[i_chan]

                    if a_chan >= 0 and a_chan < n_imag_chan:
                        u = -uvw[i_time, i_baseline, 0] * uv_scale[0, i_chan]
                        v = -uvw[i_time, i_baseline, 1] * uv_scale[1, i_chan]
                        u_pos = u + uv_center[0]
                        v_pos = v + uv_center[1]
                        u_center_indx = int(u_pos + 0.5)
                        v_center_indx = int(v_pos + 0.5)

                        if u_center_indx < n_u and v_center_indx < n_v and u_center_indx >= 0 and v_center_indx >= 0:
                            u_offset = u_center_indx - u_pos
                            u_center_offset_indx = math.floor(u_offset * oversampling + 0.5)
                            v_offset = v_center_indx - v_pos
                            v_center_offset_indx = math.floor(v_offset * oversampling + 0.5)

                            for i_pol in range(n_pol):
                                if weight[i_time, i_baseline, i_pol] != 0.0:
                                    a_pol = pol_map[i_pol]
                                    weigted_data = vis_data[i_time, i_baseline, i_chan, i_pol] * weight[
                                        i_time, i_baseline, i_pol]
                                    norm = 0.0

                                    if flag[i_time, i_baseline, i_chan, i_pol] == 0:
                                        for i_v in range(-support_center, support_center + 1):
                                            v_indx = v_center_indx + i_v
                                            v_offset_indx = np.abs(oversampling * i_v + v_center_offset_indx)
                                            conv_v = cgk_1D[v_offset_indx]

                                            for i_u in range(-support_center, support_center + 1):
                                                u_indx = u_center_indx + i_u
                                                u_offset_indx = np.abs(oversampling * i_u + u_center_offset_indx)
                                                conv_u = cgk_1D[u_offset_indx]
                                                conv = conv_u * conv_v

                                                grid[a_chan, a_pol, u_indx, v_indx] = grid[
                                                                                          a_chan, a_pol, u_indx, v_indx] + conv * weigted_data
                                                norm = norm + conv

                                    sum_weight[a_chan, a_pol] = sum_weight[a_chan, a_pol] + weight[
                                        i_time, i_baseline, i_pol] * norm

    return



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
from numba import jit
import math


def _coordinates(npixel: int):
    """ 1D array which spans [-.5,.5[ with 0 at position npixel/2
    """
    return (np.arange(npixel) - npixel // 2) / npixel


def _coordinates2(npixel: int):
    """Two dimensional grids of coordinates spanning -1 to 1 in each dimension
    1. a step size of 2/npixel and
    2. (0,0) at pixel (floor(n/2),floor(n/2))
    """
    return (np.mgrid[0:npixel, 0:npixel] - npixel // 2) / npixel


def create_prolate_spheroidal_kernel(oversampling, support, n_uv):
    """
    Create PSWF to serve as gridding kernel

    Parameters
    ----------
    oversampling : int 
        oversampling//2 is the index of the zero value of the oversampling value
    support : int
        support//2 is the index of the zero value of the support values
    n_uv: int array
        (2)
        number of pixels in u,v space

    Returns
    -------
    kernel : numpy.ndarray
    kernel_image : numpy.ndarray

    """
    # support//2 is the index of the zero value of the support values
    # oversampling//2 is the index of the zero value of the oversampling value
    support_center = support // 2
    oversampling_center = oversampling // 2

    support_values = (np.arange(support) - support_center)
    if (oversampling % 2) == 0:
        oversampling_values = ((np.arange(oversampling + 1) - oversampling_center) / oversampling)[:, None]
        kernel_points_1D = (np.broadcast_to(support_values, (oversampling + 1, support)) + oversampling_values)
    else:
        oversampling_values = ((np.arange(oversampling) - oversampling_center) / oversampling)[:, None]
        kernel_points_1D = (np.broadcast_to(support_values, (oversampling, support)) + oversampling_values)

    kernel_points_1D = kernel_points_1D / support_center

    _, kernel_1D = prolate_spheroidal_function(kernel_points_1D)
    # kernel_1D /= np.sum(np.real(kernel_1D[oversampling_center,:]))

    if (oversampling % 2) == 0:
        kernel = np.zeros((oversampling + 1, oversampling + 1, support, support),
                          dtype=np.double)  # dtype=np.complex128
    else:
        kernel = np.zeros((oversampling, oversampling, support, support), dtype=np.double)

    for x in range(oversampling):
        for y in range(oversampling):
            kernel[x, y, :, :] = np.outer(kernel_1D[x, :], kernel_1D[y, :])

    # norm = np.sum(np.real(kernel))
    # kernel /= norm

    # Gridding correction function (applied after dirty image is created)
    kernel_image_points_1D_u = np.abs(2.0 * _coordinates(n_uv[0]))
    kernel_image_1D_u = prolate_spheroidal_function(kernel_image_points_1D_u)[0]

    kernel_image_points_1D_v = np.abs(2.0 * _coordinates(n_uv[1]))
    kernel_image_1D_v = prolate_spheroidal_function(kernel_image_points_1D_v)[0]

    kernel_image = np.outer(kernel_image_1D_u, kernel_image_1D_v)

    # kernel_image[kernel_image > 0.0] = kernel_image.max() / kernel_image[kernel_image > 0.0]

    # kernel_image =  kernel_image/kernel_image.max()
    return kernel, kernel_image


def prolate_spheroidal_function(u):
    """
    Calculate PSWF using an old SDE routine re-written in Python

    Find Spheroidal function with M = 6, alpha = 1 using the rational
    approximations discussed by Fred Schwab in 'Indirect Imaging'.

    This routine was checked against Fred's SPHFN routine, and agreed
    to about the 7th significant digit.

    The griddata function is (1-NU**2)*GRDSF(NU) where NU is the distance
    to the edge. The grid correction function is just 1/GRDSF(NU) where NU
    is now the distance to the edge of the image.
    """
    p = np.array([[8.203343e-2, -3.644705e-1, 6.278660e-1, -5.335581e-1, 2.312756e-1],
                  [4.028559e-3, -3.697768e-2, 1.021332e-1, -1.201436e-1, 6.412774e-2]])
    q = np.array([[1.0000000e0, 8.212018e-1, 2.078043e-1], [1.0000000e0, 9.599102e-1, 2.918724e-1]])

    _, n_p = p.shape
    _, n_q = q.shape

    u = np.abs(u)
    uend = np.zeros(u.shape, dtype=np.float64)
    part = np.zeros(u.shape, dtype=np.int64)

    part[(u >= 0.0) & (u < 0.75)] = 0
    part[(u >= 0.75) & (u <= 1.0)] = 1
    uend[(u >= 0.0) & (u < 0.75)] = 0.75
    uend[(u >= 0.75) & (u <= 1.0)] = 1.0

    delusq = u ** 2 - uend ** 2

    top = p[part, 0]
    for k in range(1, n_p):  # small constant size loop
        top += p[part, k] * np.power(delusq, k)

    bot = q[part, 0]
    for k in range(1, n_q):  # small constant size loop
        bot += q[part, k] * np.power(delusq, k)

    grdsf = np.zeros(u.shape, dtype=np.float64)
    ok = (bot > 0.0)
    grdsf[ok] = top[ok] / bot[ok]
    ok = np.abs(u > 1.0)
    grdsf[ok] = 0.0

    # Return the griddata function and the grid correction function
    return grdsf, (1 - u ** 2) * grdsf


def create_prolate_spheroidal_kernel_1D(oversampling, support):
    support_center = support // 2
    oversampling_center = oversampling // 2
    u = np.arange(oversampling * (support_center)) / (support_center * oversampling)

    long_half_kernel_1D = np.zeros(oversampling * (support_center + 1))
    _, long_half_kernel_1D[0:oversampling * (support_center)] = prolate_spheroidal_function(u)
    return long_half_kernel_1D



from setuptools import setup, find_packages

with open('README.md', "r") as fid:   #encoding='utf-8'
    long_description = fid.read()

setup(
    name='ngcasa',
    version='0.0.1',
    description='Next Generation CASA',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='National Radio Astronomy Observatory',
    author_email='casa-feedback@nrao.edu',
    url='https://github.com/casangi/ngcasa',
    license='Apache-2.0',
    packages=find_packages(),
    install_requires=['numpy==1.18.1',
                      'numba>=0.48.0',
                      'dask>=2.10.0',
                      'bokeh>=1.4.0',
                      'pandas>=0.25.2',
                      'xarray>=0.14.1',
                      'zarr>=2.4.0',
                      'numcodecs>=0.6.4',
                      'matplotlib>=3.1.2',
                      'sparse>=0.9.1'],
    extras_require={
        'dev': [
            'pylint>=2.4.4'
            #'black>=19.10.b0',
            'pytest>=5.3.5',
            #'pytest-pep8',
            #'pytest-cov'
            's3fs>=0.4.0',
        ]
    }

)



##################################
# Non-Uniform FFT Functions
#
#
##################################
from .calc_image_cell_size import calc_image_cell_size
from .dirty_image import dirty_image
from .grid import grid
from .gridding_convolutional_kernels import create_prolate_spheroidal_kernel, create_prolate_spheroidal_kernel_1D, \
    prolate_spheroidal_function
from .vectorized_gridder import *
